/* 
 * ORACLE9 extension for IP connection tracking.
 * framework copy from ftp NAT conntrack helper.
 * written by TripleX Chong, Goldenhope Inc.
 */
#include <linux/config.h>
#include <linux/module.h>
#include <linux/netfilter.h>
#include <linux/ip.h>
#include <linux/ctype.h>
#include <net/checksum.h>
#include <net/tcp.h>

#include <linux/netfilter_ipv4/lockhelp.h>
#include <linux/netfilter_ipv4/ip_conntrack_helper.h>
#include <linux/netfilter_ipv4/ip_conntrack_oracle.h>

DECLARE_LOCK(ip_oracle_lock);
struct module *ip_conntrack_oracle = THIS_MODULE;

#if 0
#define DEBUGP printk
#else
#define DEBUGP(format, args...)
#endif

#define isascii(c) (((unsigned char)(c))<=0x7f)
#define LF 0xA
typedef uint32_t in_addr_t;

int
my_inet_aton(char *cp, u_int32_t *the_addr) {
	static const in_addr_t max[4] = { 0xffffffff, 0xffffff, 0xffff, 0xff };
	in_addr_t val;
	char c;
	union iaddr {
	  uint8_t bytes[4];
	  uint32_t word;
	} res;
	uint8_t *pp = res.bytes;
	int digit,base;

	res.word = 0;

	c = *cp;
	for (;;) {
		/*
		 * Collect number up to ``.''.
		 * Values are specified as for C:
		 * 0x=hex, 0=octal, isdigit=decimal.
		 */
		if (!isdigit(c)) goto ret_0;
		val = 0; base = 10; digit = 0;
		for (;;) {
			if (isdigit(c)) {
				val = (val * base) + (c - '0');
				c = *++cp;
				digit = 1;
			} else {
				break;
			}
		}
		if (c == '.') {
			/*
			 * Internet format:
			 *	a.b.c.d
			 *	a.b.c	(with c treated as 16 bits)
			 *	a.b	(with b treated as 24 bits)
			 */
			if (pp > res.bytes + 2 || val > 0xff) {
				goto ret_0;
			}
			*pp++ = val;
			c = *++cp;
		} else
			break;
	}
	/*
	 * Check for trailing characters.
	 */
	if (c != '\0' && (!isascii(c) || !isspace(c))) {
		goto ret_0;
	}
	/*
	 * Did we get a valid digit?
	 */
	if (!digit) {
		goto ret_0;
	}

	/* Check whether the last part is in its limits depending on
	   the number of parts in total.  */
	if (val > max[pp - res.bytes]) {
		goto ret_0;
	}

	if (the_addr != NULL) {
		*the_addr = res.word | htonl (val);
	}

	return (1);

ret_0:
	return (0);
}

static int my_atoi(const char *name)
{
    int val = 0;

    for (;; name++) {
	switch (*name) {
	    case '0'...'9':
		val = 10*val+(*name-'0');
		break;
	    default:
		return val;
	}
    }
}

int parse_ip(char *ipstr, u_int32_t *ip)
{
	return my_inet_aton(ipstr, ip);
}

int parse_port(char *portstr, u_int16_t *port)
{
	int i;
	
	i = my_atoi(portstr);
	if(i<=0 || i>65535){
		DEBUGP("conntrack_oracle: %d is not a valid port\n", i);
		return 0;
	}
	*port = i;
	*port = htons(*port);
	DEBUGP("conntrack_oracle: %d is a valid port\n", i);
	return 1;
}

/* Return 1 for match, 0 for accept, -1 for partial. */
static int find_pattern(const char *data, size_t dlen,
			unsigned int *numoff,
			unsigned int *numlen,
			u_int32_t *ip, u_int16_t *port)
{
	int i, j, k, off, level;
	char *ipstr=NULL, *portstr=NULL;
	
	if(dlen < 30) {
		DEBUGP("conntrack_oracle: dlen=%d is too short\n", dlen);
		return 0;
	}
	for(off=0; off < dlen; off++) {
		if(data[off] == '('){
			if(dlen - off <= 7){
				DEBUGP("conntrack_oracle: ADDRESS not found \n");
				return 0;
			}
			if(memcmp(&data[off+1], "ADDRESS", 7) == 0){
				DEBUGP("conntrack_oracle: ADDRESS found \n");
				break;
			}
		}
	}
	if(off >= dlen - 1) {
		DEBUGP("conntrack_oracle: ADDRESS not found \n");
		return 0;
	}
	*numoff = off - 2;   //include 2 byte length
	
	level = 0;
	for(i=off; i<dlen; i++) {
		if(data[i] == '(')
			level ++;
		if(data[i] == ')')
			level --;
		if(level == 0)
			break;
	}
	if(level != 0 ) {
		DEBUGP("conntrack_oracle: () not matched\n");
		goto nomatch;
	} else {
		DEBUGP("conntrack_oracle: () matched\n");
	}
	*numlen = i - off + 1 + 2;   //include 2 byte length

	level = 0;
	for(i=off; i<dlen; i++) {
		if(data[i] == '('){
			if(memcmp(&data[i+1], "HOST=", 5) == 0) {
				ipstr = kmalloc(20, GFP_ATOMIC);
				if(ipstr == NULL)
					goto nomatch;
				memset(ipstr, 0, 20);
				j = i+6;
				for(k=j; k<dlen; k++){
					if(data[k] == ')') {
						break;
					}
					if(k-j >= 20)
						goto nomatch;
				}
				memcpy(ipstr, &data[j], k-j);
				DEBUGP("oracle_help: parse ip %s\n", ipstr);
				if(!parse_ip(ipstr, ip))
					goto nomatch;
				level = 1;
				break;
			}
		}
	}
	if(level == 0)
		goto nomatch;
	level = 0;
	for(i=off; i<dlen; i++) {
		if(data[i] == '('){
			if(memcmp(&data[i+1], "PORT=", 5) == 0){
				portstr = kmalloc(6, GFP_ATOMIC);
				if(portstr == NULL)
					goto nomatch;
				memset(portstr, 0, 6);
				j = i+6;
				for(k=j; k<dlen; k++){
					if(data[k] == ')') {
						break;
					}
					if(k-j >= 6)
						goto nomatch;
				}
				memcpy(portstr, &data[j], k-j);
				DEBUGP("oracle_help: parse port %s\n", portstr);
				if(!parse_port(portstr, port))
					goto nomatch;
				level = 1;
				break;
			}
		}
	}
	if(level == 0)
		goto nomatch;

	DEBUGP("Match succeeded!\n");
	if(ipstr != NULL)
		kfree(ipstr);
	if(portstr != NULL)
		kfree(portstr);
	return 1;
nomatch:
	DEBUGP("conntrack_oracle: nomatch\n");
	if(ipstr != NULL)
		kfree(ipstr);
	if(portstr != NULL)
		kfree(portstr);
	return 0;
}

/* FIXME: This should be in userspace.  Later. */
static int oracle_help(const struct iphdr *iph, size_t len,
		struct ip_conntrack *ct,
		enum ip_conntrack_info ctinfo)
{
	/* tcplen not negative guaranteed by ip_conntrack_tcp.c */
	struct tcphdr *tcph = (void *)iph + iph->ihl * 4;
	const char *data = (const char *)tcph + tcph->doff * 4;
	unsigned int tcplen = len - iph->ihl * 4;
	unsigned int datalen = tcplen - tcph->doff * 4;
	u_int32_t old_seq_aft_nl;
	int old_seq_aft_nl_set;
	u_int32_t ip;
	u_int16_t port;
	int dir = CTINFO2DIR(ctinfo);
	unsigned int matchlen, matchoff;
	struct ip_conntrack_expect expect, *exp = &expect;
	struct ip_ct_oracle_expect *exp_oracle_info = &exp->help.exp_oracle_info;

	int found = 0;

	if(ctinfo != IP_CT_ESTABLISHED+IP_CT_IS_REPLY) {
		return NF_ACCEPT;
	}

	/* Not whole TCP header? */
	if (tcplen < sizeof(struct tcphdr) || tcplen < tcph->doff*4) {
		DEBUGP("oracle: tcplen = %u\n", (unsigned)tcplen);
		return NF_ACCEPT;
	}

	/* Checksum invalid?  Ignore. */
	/* FIXME: Source route IP option packets --RR */
	if (tcp_v4_check(tcph, tcplen, iph->saddr, iph->daddr,
			 csum_partial((char *)tcph, tcplen, 0))) {
		DEBUGP("oracle_help: bad csum: %p %u %u.%u.%u.%u %u.%u.%u.%u\n",
		       tcph, tcplen, NIPQUAD(iph->saddr),
		       NIPQUAD(iph->daddr));
		return NF_ACCEPT;
	}

	DEBUGP("conntrack_oracle: datalen %u\n", datalen);

	found = find_pattern(data, datalen,
			     &matchoff, &matchlen, &ip, &port);

	if (found == -1) {
		/* We don't usually drop packets.  After all, this is
		   connection tracking, not packet filtering.
		   However, it is neccessary for accurate tracking in
		   this case. */
		if (net_ratelimit())
			printk("conntrack_oracle: partial %s %u+%u\n",
			       "ADDRESS",
			       ntohl(tcph->seq), datalen);
		return NF_DROP;
	} else if (found == 0) {
		/* No match */
		DEBUGP("conntrack_oracle: pattern not found\n");
		return NF_ACCEPT;
	}

	DEBUGP("conntrack_oracle: match `%.*s' (%u bytes at %u)\n",
	       (int)matchlen, data + matchoff,
	       matchlen, ntohl(tcph->seq) + matchoff);
	       
	memset(&expect, 0, sizeof(expect));

	/* Update the oracle info */
	LOCK_BH(&ip_oracle_lock);
	exp->seq = ntohl(tcph->seq) + matchoff;
	exp_oracle_info->len = matchlen;
	exp_oracle_info->port = port;

	exp->tuple.src.ip = ct->tuplehash[!dir].tuple.src.ip;
	exp->tuple.dst.ip = ip;
	exp->tuple.dst.u.tcp.port = port;
	exp->tuple.dst.protonum = IPPROTO_TCP;
	exp->mask.src.ip = 0xFFFFFFFF;
	exp->mask.dst.ip = 0xFFFFFFFF;
	exp->mask.dst.u.tcp.port = 0xFFFF;
	exp->mask.dst.protonum = 0xFFFF;

	exp->expectfn = NULL;

	/* Ignore failure; should only happen with NAT */
	ip_conntrack_expect_related(ct, &expect);

	UNLOCK_BH(&ip_oracle_lock);

	return NF_ACCEPT;
}

static struct ip_conntrack_helper oracle;

/* Not __exit: called from init() */
static void fini(void)
{
	ip_conntrack_helper_unregister(&oracle);
}

static int __init init(void)
{
	int ret;

	memset(&oracle, 0, sizeof(struct ip_conntrack_helper));
	oracle.tuple.src.u.tcp.port = htons(ORACLE_PORT);
	oracle.tuple.dst.protonum = IPPROTO_TCP;
	oracle.mask.src.u.tcp.port = 0xFFFF;
	oracle.mask.dst.protonum = 0xFFFF;
	oracle.max_expected = 1;
	oracle.timeout = 0;
	oracle.flags = IP_CT_HELPER_F_REUSE_EXPECT;
	oracle.me = ip_conntrack_oracle;
	oracle.help = oracle_help;

	ret = ip_conntrack_helper_register(&oracle);

	if (ret) {
		fini();
		return ret;
	}

	return 0;
}

#ifdef CONFIG_IP_NF_NAT_NEEDED
EXPORT_SYMBOL(ip_oracle_lock);
#endif

MODULE_LICENSE("GPL");
module_init(init);
module_exit(fini);
