/* 
 * ORACLE9 extension for TCP NAT alteration. 
 * framework copy from ftp NAT helper.
 * written by TripleX Chong, Goldenhope Inc.
 */
#include <linux/module.h>
#include <linux/netfilter_ipv4.h>
#include <linux/ip.h>
#include <linux/tcp.h>
#include <net/tcp.h>
#include <linux/netfilter_ipv4/ip_nat.h>
#include <linux/netfilter_ipv4/ip_nat_helper.h>
#include <linux/netfilter_ipv4/ip_nat_rule.h>
#include <linux/netfilter_ipv4/ip_conntrack_oracle.h>
#include <linux/netfilter_ipv4/ip_conntrack_helper.h>

#if 0
#define DEBUGP printk
#else
#define DEBUGP(format, args...)
#endif

DECLARE_LOCK_EXTERN(ip_oracle_lock);

/* FIXME: Time out? --RR */

static unsigned int
oracle_nat_expected(struct sk_buff **pskb,
		 unsigned int hooknum,
		 struct ip_conntrack *ct,
		 struct ip_nat_info *info)
{
	struct ip_nat_multi_range mr;
	u_int32_t newdstip, newsrcip, newip;
	struct ip_ct_oracle_expect *exp_oracle_info;

	struct ip_conntrack *master = master_ct(ct);
	
	IP_NF_ASSERT(info);
	IP_NF_ASSERT(master);

	IP_NF_ASSERT(!(info->initialized & (1<<HOOK2MANIP(hooknum))));

	DEBUGP("nat_expected: We have a connection!\n");
	exp_oracle_info = &ct->master->help.exp_oracle_info;

	LOCK_BH(&ip_oracle_lock);

	/* make the connection go to the server */
	newdstip = master->tuplehash[IP_CT_DIR_REPLY].tuple.src.ip;
	newsrcip = master->tuplehash[IP_CT_DIR_REPLY].tuple.dst.ip;
	DEBUGP("nat_expected: PASV cmd. %u.%u.%u.%u->%u.%u.%u.%u\n",
	       NIPQUAD(newsrcip), NIPQUAD(newdstip));

	UNLOCK_BH(&ip_oracle_lock);

	if (HOOK2MANIP(hooknum) == IP_NAT_MANIP_SRC)
		newip = newsrcip;
	else
		newip = newdstip;

	DEBUGP("nat_expected: IP to %u.%u.%u.%u\n", NIPQUAD(newip));

	mr.rangesize = 1;
	/* We don't want to manip the per-protocol, just the IPs... */
	mr.range[0].flags = IP_NAT_RANGE_MAP_IPS;
	mr.range[0].min_ip = mr.range[0].max_ip = newip;

	/* ... unless we're doing a MANIP_DST, in which case, make
	   sure we map to the correct port */
	if (HOOK2MANIP(hooknum) == IP_NAT_MANIP_DST) {
		mr.range[0].flags |= IP_NAT_RANGE_PROTO_SPECIFIED;
		mr.range[0].min = mr.range[0].max
			= ((union ip_conntrack_manip_proto)
				{ exp_oracle_info->port });
	}
	return ip_nat_setup_info(ct, &mr, hooknum);
}


/* |1|132.235.1.2|6275| */
static int
mangle_oracle_packet(struct sk_buff **pskb,
		   u_int32_t newip,
		   u_int16_t port,
		   unsigned int matchoff,
		   unsigned int matchlen,
		   struct ip_conntrack *ct,
		   enum ip_conntrack_info ctinfo)
{
	int ret = 0;
	char *buffer;
	u_int16_t *l, bl;
	int size = sizeof("(ADDRESS=(PROTOCOL=tcp)(HOST=255.255.255.255)(PORT=65535))") + 2;

	MUST_BE_LOCKED(&ip_oracle_lock);
	
	buffer = kmalloc(size, GFP_ATOMIC);
	if(buffer == NULL)
		return 0;
	memset(buffer, 0, size);
	sprintf(buffer+2, "(ADDRESS=(PROTOCOL=tcp)(HOST=%u.%u.%u.%u)(PORT=%d))", NIPQUAD(newip), ntohs(port));
	l = (u_int16_t *)buffer;
	bl = strlen(buffer+2);
	if(bl < matchlen)
		bl = matchlen;  /* note: the packet can not be shorter than origin packet*/
	*l = htons(bl);
	DEBUGP("calling ip_nat_mangle_tcp_packet\n");
	DEBUGP("new buffer = %s \n", buffer+2);

	ret = ip_nat_mangle_tcp_packet(pskb, ct, ctinfo, matchoff, 
					matchlen, buffer, ntohs(*l)+2);
	kfree(buffer);
	return ret;
}

static int oracle_data_fixup(const struct ip_ct_oracle_expect *ct_oracle_info,
			  struct ip_conntrack *ct,
			  struct sk_buff **pskb,
			  enum ip_conntrack_info ctinfo,
			  struct ip_conntrack_expect *expect)
{
	u_int32_t newip;
	struct iphdr *iph = (*pskb)->nh.iph;
	struct tcphdr *tcph = (void *)iph + iph->ihl*4;
	u_int16_t port;
	struct ip_conntrack_tuple newtuple;

	MUST_BE_LOCKED(&ip_oracle_lock);
	DEBUGP("ORACLE_NAT: seq %u + %u in %u\n",
	       expect->seq, ct_oracle_info->len,
	       ntohl(tcph->seq));

	/* Change address inside packet to match way we're mapping
	   this connection. */
	newip = ct->tuplehash[IP_CT_DIR_ORIGINAL].tuple.dst.ip;
	/* Expect something from client->server */
	newtuple.src.ip = 
		ct->tuplehash[IP_CT_DIR_ORIGINAL].tuple.src.ip;
	newtuple.dst.ip = 
		ct->tuplehash[IP_CT_DIR_ORIGINAL].tuple.dst.ip;
	newtuple.dst.protonum = IPPROTO_TCP;
	newtuple.src.u.tcp.port = expect->tuple.src.u.tcp.port;

	/* Try to get same port: if not, try to change it. */
	for (port = ct_oracle_info->port; port != 0; port++) {
		newtuple.dst.u.tcp.port = port;

		if (ip_conntrack_change_expect(expect, &newtuple) == 0)
			break;
	}
	if (port == 0)
		return 0;

	if (!mangle_oracle_packet(pskb, newip, port,
			  expect->seq - ntohl(tcph->seq),
			  ct_oracle_info->len, ct, ctinfo))
		return 0;

	return 1;
}

static unsigned int oracle_help(struct ip_conntrack *ct,
			 struct ip_conntrack_expect *exp,
			 struct ip_nat_info *info,
			 enum ip_conntrack_info ctinfo,
			 unsigned int hooknum,
			 struct sk_buff **pskb)
{
	struct iphdr *iph = (*pskb)->nh.iph;
	struct tcphdr *tcph = (void *)iph + iph->ihl*4;
	unsigned int datalen;
	int dir;
	struct ip_ct_oracle_expect *ct_oracle_info;

	if (!exp)
		DEBUGP("ip_nat_oracle: no exp!!");

	ct_oracle_info = &exp->help.exp_oracle_info;

	/* Only mangle things once: original direction in POST_ROUTING
	   and reply direction on PRE_ROUTING. */
	dir = CTINFO2DIR(ctinfo);
	if (!((hooknum == NF_IP_POST_ROUTING && dir == IP_CT_DIR_ORIGINAL)
	      || (hooknum == NF_IP_PRE_ROUTING && dir == IP_CT_DIR_REPLY))) {
		DEBUGP("nat_oracle: Not touching dir %s at hook %s\n",
		       dir == IP_CT_DIR_ORIGINAL ? "ORIG" : "REPLY",
		       hooknum == NF_IP_POST_ROUTING ? "POSTROUTING"
		       : hooknum == NF_IP_PRE_ROUTING ? "PREROUTING"
		       : hooknum == NF_IP_LOCAL_OUT ? "OUTPUT" : "???");
		return NF_ACCEPT;
	}

	datalen = (*pskb)->len - iph->ihl * 4 - tcph->doff * 4;
	LOCK_BH(&ip_oracle_lock);
	/* If it's in the right range... */
	if (between(exp->seq + ct_oracle_info->len,
		    ntohl(tcph->seq),
		    ntohl(tcph->seq) + datalen)) {
		if (!oracle_data_fixup(ct_oracle_info, ct, pskb, ctinfo, exp)) {
			UNLOCK_BH(&ip_oracle_lock);
			return NF_DROP;
		}
	} else {
		/* Half a match?  This means a partial retransmisison.
		   It's a cracker being funky. */
		if (net_ratelimit()) {
			printk("ORACLE_NAT: partial packet %u/%u in %u/%u\n",
			       exp->seq, ct_oracle_info->len,
			       ntohl(tcph->seq),
			       ntohl(tcph->seq) + datalen);
		}
		UNLOCK_BH(&ip_oracle_lock);
		return NF_DROP;
	}
	UNLOCK_BH(&ip_oracle_lock);

	return NF_ACCEPT;
}

static struct ip_nat_helper oracle;

/* Not __exit: called from init() */
static void fini(void)
{
	ip_nat_helper_unregister(&oracle);
}

static int __init init(void)
{
	int ret = 0;

	memset(&oracle, 0, sizeof(struct ip_nat_helper));

	oracle.tuple.dst.protonum = IPPROTO_TCP;
	oracle.tuple.src.u.tcp.port = htons(ORACLE_PORT);
	oracle.mask.dst.protonum = 0xFFFF;
	oracle.mask.src.u.tcp.port = 0xFFFF;
	oracle.help = oracle_help;
	oracle.me = THIS_MODULE;
	oracle.flags = 0;
	oracle.expect = oracle_nat_expected;

	ret = ip_nat_helper_register(&oracle);

	if (ret) {
		printk("ip_nat_oracle: error registering "
		       "helper for port %d\n", ORACLE_PORT);
		fini();
	return ret;
	}

	return ret;
}

module_init(init);
module_exit(fini);
MODULE_LICENSE("GPL");
